import inspect
import sys
import os
import pandas as pd
from pathlib import Path

from weaveio.hierarchy import Multiple, Hierarchy, OneOf, Optional
from weaveio.opr3.hierarchy import Spectrum, Author, APS, Measurement, \
    Single, FibreTarget, Exposure, Stack, OB, Superstack, \
    OBSpec, Supertarget, WeaveTarget, _predicate, MCMCMeasurement, Line, SpectralIndex, RedshiftMeasurement, Spectrum1D, ArrayHolder
from weaveio.opr3.l1 import L1Spectrum, L1SingleSpectrum, L1StackSpectrum, L1SupertargetSpectrum, L1SuperstackSpectrum, L1StackedSpectrum

HERE = Path(os.path.dirname(os.path.abspath(__file__)))
gandalf_lines = pd.read_csv(HERE / 'expected_lines.csv', sep=' ')
gandalf_indices = pd.read_csv(HERE / 'expected_line_indices.csv', sep=' ')
# gandalf_lines['name'] = gandalf_lines['name'].str.replace('[', '').str.replace(']', '').str.lower()
gandalf_line_names = (gandalf_lines['name'] + '_' + gandalf_lines['lambda'].apply(lambda x: f'{x:.2f}')).values.tolist()
gandalf_index_names = gandalf_indices['name'].values.tolist()


class L2(Hierarchy):
    is_template = True


class IngestedSpectrum(Spectrum1D):
    """
    An ingested spectrum is one which is a slightly modified version of an L1 spectrum
    """
    is_template = True
    children = []  # gets rid of inherited wvl array
    products = ['flux', 'error', 'wvl', 'ivar', 'logwvl', 'goodpix']


class UncombinedIngestedSpectrum(IngestedSpectrum):
    parents = [L1Spectrum]
    identifier_builder = ['l1_spectrum']


class CombinedIngestedSpectrum(IngestedSpectrum):
    parents = [Multiple(L1Spectrum, 1, 3)]
    identifier_builder = ['l1_spectra']


class ModelSpectrum(Spectrum1D):
    is_template = True
    children = []  # gets rid of inherited wvl array
    parents = [OneOf(IngestedSpectrum, one2one=True)]
    identifier_builder = ['ingested_spectrum']
    products = ['flux']


class UncombinedModelSpectrum(ModelSpectrum):
    parents = [OneOf(UncombinedIngestedSpectrum, one2one=True)]
    identifier_builder = ['uncombined_ingested_spectrum']


class BaseCombinedModelSpectrum(ModelSpectrum):
    is_template = True
    parents = [OneOf(CombinedIngestedSpectrum, one2one=True)]
    identifier_builder = ['combined_ingested_spectrum']


class CombinedModelSpectrum(BaseCombinedModelSpectrum):
    is_template = False


# This allows us to use 'clean' to talk about the clean model or clean spectrum
# IngestedSpectrum->ModelSpectrum->CleanModelSpectrum
# IngestedSpectrum->CleanIngestedSpectrum

class GandalfSpectrum(ModelSpectrum):
    """template object that only stores model flux and sources the wavelength array from another linked object"""
    is_template = True
    parents = []  # removes input spectrum dependency
    identifier_builder = None


class GandalfModelSpectrum(BaseCombinedModelSpectrum, GandalfSpectrum):
    """
    A Gandalf model spectrum is the full modelled SED of emission lines and continuum
    It also has:
        - `.clean_model`     : a clean model spectrum without emission line contribution
        - `.clean_observed`  : the observed spectrum with model emission lines subtracted
        - `.emission_line_model`  : the model emission line contribution only
    """

class GandalfDerivativeSpectrum(GandalfSpectrum):
    is_template = True
    parents = [OneOf(GandalfModelSpectrum, one2one=True)]
    identifier_builder = ['gandalf_model_spectrum']


class GandalfCleanIngestedSpectrum(GandalfDerivativeSpectrum):
    pass


class GandalfCleanModelSpectrum(GandalfDerivativeSpectrum):
    pass


class GandalfEmissionModelSpectrum(GandalfDerivativeSpectrum):
    pass


class Fit(Hierarchy):
    """
    A fit is the result of applying fitting_software to an ingested spectrum
    In the case of combined spectra being available, there is only one ingested spectrum input
    otherwise, there are more.
    """
    is_template = True
    parents = [Multiple(L1Spectrum, 2, 3, one2one=True),
               Multiple(UncombinedModelSpectrum, 0, 3, one2one=True),
               Optional(BaseCombinedModelSpectrum, one2one=True),
               Multiple(ModelSpectrum, 1, 3, one2one=True, notreal=True)]


class RedshiftArray(ArrayHolder):
    factors = ['value', 'start', 'end', 'step']
    identifier_builder = ['start', 'end', 'step']


class Template(Fit):
    parents = [Multiple(L1Spectrum, 2, 3),
               Multiple(UncombinedModelSpectrum, 0, 3, one2one=True), Optional(CombinedModelSpectrum, one2one=True),
               Multiple(ModelSpectrum, 1, 3, one2one=True, notreal=True)]
    factors = ['chi2_array', 'name']
    children = [OneOf(RedshiftArray, one2one=True)]
    identifier_builder = ['uncombined_model_spectra', 'combined_model_spectrum', 'name']


class Redrock(Fit):
    factors = ['class', 'subclass', 'snr', 'chi2', 'deltachi2', 'ncoeff', 'coeff',
               'npixels', 'srvy_class', 'z', 'zerr', 'zwarn']
    parents = [Multiple(L1Spectrum, 2, 3),
               Multiple(UncombinedModelSpectrum, 0, 3, one2one=True), Optional(CombinedModelSpectrum, one2one=True),
               Multiple(ModelSpectrum, 1, 3, one2one=True, notreal=True)]
    # todo: really need to have a check in the merge that checks that combinedmodelspec doesnt exist if not given
    template_names = ['galaxy', 'qso', 'star_a', 'star_b', 'star_cv', 'star_f', 'star_g', 'star_k', 'star_m', 'star_wd']
    parents += [OneOf(Template, idname=x, one2one=True) for x in template_names]
    identifier_builder = ['uncombined_model_spectra', 'combined_model_spectrum']


class RVSpecfit(Fit):
    singular_name = 'rvspecfit'
    parents = [Multiple(L1Spectrum, 2, 3),
               Multiple(UncombinedModelSpectrum, 0, 3, one2one=True), Optional(CombinedModelSpectrum, one2one=True),
               Multiple(ModelSpectrum, 1, 3, one2one=True, notreal=True)]
    factors = Fit.factors + ['skewness', 'kurtosis', 'vsini', 'snr', 'chi2_tot']
    factors += Measurement.as_factors('vrad', 'logg', 'teff', 'feh', 'alpha')
    identifier_builder = ['uncombined_model_spectra', 'combined_model_spectrum']


class Ferre(Fit):
    parents = [Multiple(L1Spectrum, 2, 3),
               Multiple(UncombinedModelSpectrum, 0, 3, one2one=True), Optional(CombinedModelSpectrum, one2one=True),
               Multiple(ModelSpectrum, 1, 3, one2one=True, notreal=True)]
    factors = Fit.factors + ['snr', 'chi2_tot', 'flag']
    factors += Measurement.as_factors('micro', 'logg', 'teff', 'feh', 'alpha', 'elem')
    identifier_builder = ['uncombined_model_spectra', 'combined_model_spectrum']


class Gandalf(Fit):
    plural_name = 'gandalfs'
    parents = [Multiple(L1Spectrum, 2, 3),
               OneOf(GandalfModelSpectrum, one2one=True, idname='model'),
               OneOf(GandalfCleanIngestedSpectrum, one2one=True, idname='clean_observed'),
               OneOf(GandalfCleanModelSpectrum, one2one=True, idname='clean_model'),
               OneOf(GandalfEmissionModelSpectrum, one2one=True, idname='emission_line_model'),
               ]
    factors = Fit.factors + ['fwhm_flag']
    factors += Line.as_factors(*gandalf_line_names) + SpectralIndex.as_factors(*gandalf_index_names)
    identifier_builder = ['model']


class PPXF(Fit):
    parents = [Multiple(L1Spectrum, 2, 3), OneOf(CombinedModelSpectrum, one2one=True)]
    factors = Fit.factors + MCMCMeasurement.as_factors('v', 'sigma', 'h3', 'h4', 'h5', 'h6') + \
              Measurement.as_factors('zcorr')
    identifier_builder = ['combined_model_spectrum']


class L2Product(L2):
    is_template = True
    parents = [Multiple(L1Spectrum, 2, 3),
               Optional(Redrock, one2one=True), Optional(RVSpecfit, one2one=True),
               Optional(Ferre, one2one=True), Optional(PPXF, one2one=True), Optional(Gandalf, one2one=True)]


# L2 data products are formed from 2 or more L1 data products from different arms (red, blue, or green)
# L2 singles can only be formed from 2 single L1 data products
# Since an OB has a fixed instrument configuration, L2 stacks can only be formed from 2 L1 stacks
# However, APS tries to create the widest and deepest data possible, so L2 superstacks are not limit in their L1 spectra provenance

class L2Single(L2Product, Single):
    """
    An L2 data product resulting from two or sometimes three single L1 spectra.
    The L2 data products contain information generated by APS namely redshifts, emission line properties and model spectra.

    """
    singular_name = 'l2single'
    parents = L2Product.parents[1:] + [Multiple(L1SingleSpectrum, 2, 2, constrain=(FibreTarget, Exposure))]
    identifier_builder = ['l1single_spectra', 'fibre_target', 'exposure']


class L2Stack(L2Product, Stack):
    """
    An L2 data product resulting from two or sometimes three stacked/single L1 spectra.
    The L2 data products contain information generated by APS namely redshifts, emission line properties and model spectra.
    """
    singular_name = 'l2stack'
    parents = L2Product.parents[1:] + [Multiple(L1StackSpectrum, 2, 2, constrain=(FibreTarget, OB))]
    identifier_builder = ['l1stack_spectra', 'fibre_target', 'ob']


class L2Superstack(L2Product, Superstack):
    """
    An L2 data product resulting from two or sometimes three super-stacked/stacked/single L1 spectra.
    The L2 data products contain information generated by APS namely redshifts, emission line properties and model spectra.
    """
    singular_name = 'l2superstack'
    parents = L2Product.parents[1:] + [Multiple(L1StackedSpectrum, 2, 3, constrain=(FibreTarget, OBSpec))]
    identifier_builder = ['l1stacked_spectra', 'fibre_target', 'obspec']


class L2Supertarget(L2Product, Supertarget):
    """
    An L2 data product resulting from two or sometimes three supertarget L1 spectra.
    The L2 data products contain information generated by APS namely redshifts, emission line properties and model spectra.
    """
    singular_name = 'l2supertarget'
    parents = L2Product.parents[1:] + [Multiple(L1SupertargetSpectrum, 2, 3, constrain=(WeaveTarget,))]
    identifier_builder = ['l1supertarget_spectra', 'weave_target']


hierarchies = [i[-1] for i in inspect.getmembers(sys.modules[__name__], _predicate)]



from _warnings import warn
from collections import defaultdict
from typing import List, Union, Type, Tuple
from warnings import warn

import networkx as nx
import py2neo
import numpy as np

from .common import FrozenQuery, AmbiguousPathError, is_regex
from .dissociated import Dissociated
from .factor import SingleFactorFrozenQuery, TableFactorFrozenQuery, FactorFrozenQuery
from .tree import Branch
from ..helper import _convert_obj
from ..hierarchy import Hierarchy, Multiple, One2One
from ..writequery import CypherVariable


GET_PRODUCT = "[({{h}})<-[p:product {{{{name: '{name}'}}}}]-(hdu: HDU) | [hdu.sourcefile, hdu.extn, p.index, p.column_name]]"
GET_FACTOR = "{{h}}.{name}"
GET_FACTOR_FORCE_PLURAL = f"[{GET_FACTOR}]"


class HierarchyFrozenQuery(FrozenQuery):
    def __getitem__(self, item):
        raise NotImplementedError

    def __getattr__(self, item):
        raise NotImplementedError

    def _get_factor(self, item, plural):
        raise NotImplementedError(f"Getting singular factors is not supported for {self.__class__.__name__}")

    def _get_hierarchy(self, item, plural):
        raise NotImplementedError(f"Getting singular hierarchies is not supported for {self.__class__.__name__}")

    def _get_table_factor(self, item):
        raise NotImplementedError(f"Getting table factors is not supported for {self.__class__.__name__}")

    def _filter_by_identifiers(self, items):
        raise NotImplementedError(f"Filtering by multiple identifiers is not supported for {self.__class__.__name__}")

    def _filter_by_identifier(self, item):
        raise NotImplementedError(f"Filtering by an identifier is not supported for {self.__class__.__name__}")

    def _filter_by_boolean(self, condition):
        raise NotImplementedError(f"Filtering by a boolean condition is not supported for {self.__class__.__name__}")

    def _apply_aligning_func(self, string, other: 'HierarchyFrozenQuery'):
        if isinstance(other, HierarchyFrozenQuery):
            aligned = self.branch.align(other.branch)
            parent = self
            inputs = {
                'x': aligned.action.transformed_variables.get(parent.branch.current_hierarchy, parent.branch.current_hierarchy),
                'y': aligned.action.transformed_variables.get(other.branch.current_hierarchy, other.branch.current_hierarchy)
            }
        elif isinstance(other, Hierarchy):
            if other.identifier is None:
                raise ValueError(f"Cannot compare with an out-of-db object if it has no identifier")
            parent = getattr(self, self.hierarchy_type.idname)
            data = parent.branch.add_data(other.identifier)
            inputs = {
                'x': parent.branch.current_variables[0],
                'y': data.current_variables[0]
            }
            aligned = data
        elif isinstance(other, FrozenQuery):
            raise TypeError(f"Can only compare an object with another object not with {type(other)}")
        else:
            parent = getattr(self, self.hierarchy_type.idname)
            data = parent.branch.add_data(other)
            inputs = {
                'x': parent.branch.current_variables[0],
                'y': data.current_variables[0]
            }
            aligned = data
        newbranch = aligned.operate(string, **inputs)
        return Dissociated(self.handler, newbranch, newbranch.current_variables[-1], self)

    def __eq__(self, other: Union[Hierarchy, 'HierarchyFrozenQuery', int, float, str]) -> 'Dissociated':
        string = '{x} = {y}'
        if isinstance(other, str):
            if is_regex(other):
                string = '{x} =~ {y}'
                other = other.strip('/')
        return self._apply_aligning_func(string, other)

    def __ne__(self, other: Union[Hierarchy, 'HierarchyFrozenQuery', int, float, str]) -> 'Dissociated':
        string = '{x} <> {y}'
        if isinstance(other, str):
            if is_regex(other):
                string = 'NOT ({x} =~ {y})'
                other = other.strip('/')
        return self._apply_aligning_func(string, other)


class HeterogeneousHierarchyFrozenQuery(HierarchyFrozenQuery):
    """
    The start point for building queries. e.g. `data`
    Available data calls:
        single factor `data.runids` (but always plural)
        single hierarchy `data.runs` (but always plural)
    """
    executable = False

    def __repr__(self):
        return f'query("{self.data.rootdir}/")'

    def _get_hierarchy(self, hierarchy_name, plural) -> 'DefiniteHierarchyFrozenQuery':
        paths, hiers, startbase, endbase = self.handler.paths2hierarchy(hierarchy_name, plural=plural)
        new = self.branch.handler.begin(endbase.__name__)
        return DefiniteHierarchyFrozenQuery(self.handler, new, endbase, new.current_hierarchy, [], self)

    def _get_factor(self, factor_name, plural):
        factor_name = self.data.singular_name(factor_name)
        pathdict, base, is_product, factor_name = self.handler.paths2factor(factor_name, plural=plural)
        begin = self.branch.handler.begin(base.__name__)
        if is_product:
            func = GET_PRODUCT.format(name=factor_name)
        else:
            func = GET_FACTOR.format(name=factor_name)
        new = begin.operate(func, h=begin.current_hierarchy)
        return SingleFactorFrozenQuery(self.handler, new, factor_name, new.current_variables[0], is_product, self)

    def __getattr__(self, item):
        if item in self.data.plural_factors:
            return self._get_factor(item, plural=True)
        elif self.data.is_singular_name(item):
            raise AmbiguousPathError(f"Cannot return only one {item} since the DB contains more than one. Either filter your query first or try {self.data.plural_name(item)}")
        elif item in self.data.plural_hierarchies:
            name = self.data.singular_name(item)
            return self._get_hierarchy(name, plural=True)
        else:
            autosuggest(item, self.handler.data)


class DefiniteHierarchyFrozenQuery(HierarchyFrozenQuery):
    """
    The template class for hierarchy classes that are not heterogeneous i.e. they have a defined hierarchy type
    The start point for building queries. E.g. `data.obs, data.runs.exposure`
    It can have ids or not.
    Available data calls:
        single factor `data.exposures.runids`
        single hierarchy `data.runs.exposure`
        table of factors `data.runs[['runid', 'expmjd']]`
        filter by id `data.runs[11234]`
        filter by condition `data.runs[data.runs.runid > 0]`
    """
    def __init__(self, handler, branch: Branch, hierarchy_type: Type[Hierarchy], hierarchy_variable: CypherVariable,
                 identifiers: List, parent: 'FrozenQuery'):
        super().__init__(handler, branch, parent)
        self.hierarchy_type = hierarchy_type
        self.hierarchy_variable = hierarchy_variable
        self.identifiers = identifiers
        self.string = f"{self.hierarchy_type.singular_name}"

    def _filter_by_boolean(self, boolean_filter: 'FrozenQuery'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.hierarchy_type, self.hierarchy_variable, self.identifiers, self)

    def _prepare_query(self):
        """Add a hierarchy node return statement"""
        query = super(DefiniteHierarchyFrozenQuery, self)._prepare_query()
        with query:
            query.returns(self.branch.find_hierarchies()[-1])
        return query

    def _process_result_row(self, row, nodetype):
        node = row[0]
        if node is None:
            return None
        inputs = {}
        for f in nodetype.factors:
            inputs[f] = node[f]
        if nodetype.idname is not None:
            inputs[nodetype.idname] = node[nodetype.idname]
        base_query = self.handler.hierarchy_from_neo4j_identity(nodetype, node.identity)
        for p in nodetype.parents:
            if isinstance(p, One2One):
                inputs[p.singular_name] = getattr(base_query, p.plural_name)
            elif isinstance(p, Multiple):
                inputs[p.plural_name] = getattr(base_query, p.plural_name)
            else:
                try:
                    inputs[p.singular_name] = getattr(base_query, p.singular_name)
                except AmbiguousPathError:
                    inputs[p.singular_name] = getattr(base_query, p.plural_name)  # this should not have to be done
        h = nodetype(**inputs, do_not_create=True)
        h.add_parent_query(base_query)
        h.add_parent_data(self.handler.data)
        return h

    def _post_process(self, result: py2neo.database.Cursor, squeeze: bool = True):
        result = result.to_table()
        if len(result) == 1 and result[0] is None:
            return []
        results = []
        for row in result:
            h = self._process_result_row(row, self.hierarchy_type)
            results.append(h)
        if len(results) == 1 and squeeze:
            return results[0]
        return results

    def _get_hierarchy(self, name, plural):
        pathlist, endlist, starthier, endhier = self.handler.paths2hierarchy(name, plural=plural, start=self.hierarchy_type)
        new = self.branch.traverse(*pathlist)
        return DefiniteHierarchyFrozenQuery(self.handler, new, endhier, new.current_hierarchy, [], self)

    def _get_factor_query_paths(self, names, plurals):
        if not isinstance(names, (list, tuple)):
            names = [names]
        if not isinstance(plurals, (list, tuple)):
            plurals = [plurals]
        names = [self.data.singular_name(name) for name in names]
        local = []
        remote = defaultdict(list)
        remote_paths = {}
        for i, (name, plural) in enumerate(zip(names, plurals)):
            pathsdict, basehier, is_product, name = self.handler.paths2factor(name, plural, self.hierarchy_type)
            names[i] = name
            if basehier == self.hierarchy_type:
                local.append((name, plural, is_product))
            else:
                remote[(basehier, plural)].append((name, is_product))
                remote_paths[(basehier, plural)] = {path for pathset in pathsdict.values() for path in pathset}
        return names, plurals, local, remote, remote_paths

    def _get_multifactor_query(self, names: Union[List[str], str], plurals: Union[List[bool], bool],
                               collect_plurals: bool = True) -> Tuple[Branch, List[CypherVariable], List[bool]]:
        """
        Return the query branch, variables of a list of factor/product names
        We do this by grouping into the containing hierarchies and traversing each branch before collapsing
        returns:
                branch: The new branch object to continue the query with
                variables: The list of CypherVariable that contains the query result
                is_product: The list of True if the factor is a product
        """
        # TODO: tidy this all up
        names, plurals, local, remote, remote_paths = self._get_factor_query_paths(names, plurals)
        variables = {}
        is_products = {}
        branch = self.branch

        for (basehier, plural), factor_product_tuples in remote.items():
            paths = remote_paths[(basehier, plural)]
            travel = branch.traverse(*paths)
            funcs = []
            for name, is_product in factor_product_tuples:
                if is_product:
                    func = GET_PRODUCT.format(name=name)
                else:
                    func = GET_FACTOR.format(name=name)
                funcs.append(func)
            operate = travel.operate(*funcs, h=travel.current_hierarchy)
            if collect_plurals:
                if plural:
                    branch = branch.collect([], [operate])
                else:
                    branch = branch.collect([operate], [])
                for v, (name, is_product) in zip(operate.current_variables, factor_product_tuples):
                    variables[(name, plural)] = branch.action.transformed_variables[v]
                    is_products[(name, plural)] = is_product
            else:
                branch = operate
                for v, (name, is_product) in zip(operate.current_variables, factor_product_tuples):
                    variables[(name, plural)] = v
                    is_products[(name, plural)] = is_product

        if len(local):
            funcs = []
            for name, plural, is_product in local:
                if is_product:
                    func = GET_PRODUCT.format(name=name)
                elif plural and collect_plurals:
                    func = GET_FACTOR_FORCE_PLURAL.format(name=name)
                else:
                    func = GET_FACTOR.format(name=name)
                funcs.append(func)
            branch = branch.operate(*funcs, h=self.hierarchy_variable)
            for v, (k, plural, is_product) in zip(branch.action.output_variables, local):
                variables[(k, plural)] = v
                is_products[(k, plural)] = is_product

        # now propagate variables forward
        values = [variables[(name, plural)] for name, plural in zip(names, plurals)]
        variables = branch.get_variables(values)
        return branch, variables, [is_products[(name, plural)] for name, plural in zip(names, plurals)]

    def _get_factor(self, name, plural):
        branch, factor_variables, is_products = self._get_multifactor_query([name], [plural], collect_plurals=False)
        return SingleFactorFrozenQuery(self.handler, branch, name, factor_variables[0], is_products[0], self)

    def _get_factor_table_query(self, item) -> FactorFrozenQuery:
        """
        __getitem__ is for returning factors and ids
        There are three types of getitem input values:
        List: [[a, b]], where labelled table-like rows are output
        Tuple: [a, b], where a list of unlabelled dictionaries are output
        str: [a], where a single value is returned

        In all three cases, you still need to specify plural or singular forms.
        This allows you to have a row of n dimensional heterogeneous data.
        returns query and the labels (if any) for the table
        """
        if isinstance(item, tuple):  # return without headers
            return_keys = list(item)
            keys = list(item)
        elif isinstance(item, list):
            keys = item
            return_keys = item
        elif item is None:
            raise TypeError("item must be of type list, tuple, or str")
        else:
            raise KeyError(f"Unknown item {item} for `{self}`")
        plurals = [not self.data.is_singular_name(i) for i in item]
        branch, factor_variables, is_products = self._get_multifactor_query(keys, plurals, collect_plurals=len(plurals) > 0)
        if len(factor_variables) == 1:
            return SingleFactorFrozenQuery(self.handler, branch, keys[0], factor_variables[0], is_products[0], self.parent)
        return TableFactorFrozenQuery(self.handler, branch, keys, factor_variables, plurals, is_products, return_keys, self.parent)

    def _filter_by_identifiers(self, identifiers: List[Union[str, int, float]]) -> 'DefiniteHierarchyFrozenQuery':
        idname = self.hierarchy_type.idname
        if isinstance(identifiers, np.ndarray):
            identifiers = identifiers.ravel()
        new = self.branch.add_data(identifiers)
        identifiers_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' in {identifiers}', h=self.hierarchy_variable, identifiers=identifiers_var)
        return DefiniteHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, identifiers, self)

    def _filter_by_identifier(self, identifier: Union[str, int, float]) -> 'DefiniteHierarchyFrozenQuery':
        idname = self.hierarchy_type.idname
        new = self.branch.add_data(identifier)
        identifier_var = new.current_variables[0]
        branch = new.filter('{h}.' + idname + ' = {identifier}', h=self.hierarchy_variable, identifier=identifier_var)
        return DefiniteHierarchyFrozenQuery(self.handler, branch, self.hierarchy_type, self.hierarchy_variable, [identifier], self)

    def __getitem__(self, item):
        """
        If there is a scalar, then return a SingleFactorFrozenQuery otherwise assume its a request for a table
        The order of resolution is:
            vector hierarchy/
        """
        if isinstance(item, Dissociated):
            return self._filter_by_boolean(item)
        if isinstance(item, (list, tuple, np.ndarray)):
            valids = list(map(self.data.is_valid_name, item))
            if all(valids):
                return self._get_factor_table_query(item)
            elif any(valids):
                bad = [i for i, v in zip(item, valids) if not v]
                raise KeyError(f"{bad} are not recognised")
            else:
                return self._filter_by_identifiers(item)
        if not self.data.is_valid_name(item):
            return self._filter_by_identifier(item)  # then assume its an ID
        if isinstance(item, str) and self.data.is_valid_name(item):
            return self._get_factor_table_query([item])
        else:
            return getattr(self, item)

    def __getattr__(self, item):
        plural = self.data.is_plural_name(item)
        factor = self.data.is_factor_name(item)
        exists = self.data.is_singular_name(item) or plural
        if not exists:
            autosuggest(item, self.handler.data, self.hierarchy_type)
        if factor:
            return self._get_factor(item, plural=plural)
        else:
            return self._get_hierarchy(item, plural=plural)


def _autosuggest(a, data, relative_to=None):
    import textdistance
    distance, distance_reverse = textdistance.jaro_winkler, True
    suggestions = []
    if relative_to is not None:
        relative_to, data = _convert_obj(relative_to, data)

    for h in data.hierarchies:
        newsuggestions = []
        if relative_to is not None:
            try:
                data.find_hierarchy_paths(relative_to, h, plural=False)
            except (nx.NetworkXNoPath, AmbiguousPathError):
                plural = True
                hier = h.plural_name
                factors = [data.plural_name(f) for f in h.products_and_factors]
            else:
                plural = False
                hier = h.singular_name
                factors = h.products_and_factors
            newsuggestions.append(hier)
            newsuggestions += factors
        else:
            newsuggestions += [h.singular_name, h.plural_name] + h.products_and_factors
            newsuggestions += [data.plural_name(f) for f in h.products_and_factors]
        try:
            newsuggestions.index(a)
        except ValueError:
            suggestions += newsuggestions
        else:
            return [a]
    inorder = sorted(list(set(suggestions)), key=lambda x: distance(a, x), reverse=distance_reverse)
    return inorder[:3]


def autosuggest(a, data, relative_to=None):
    suffix = '\nYou can learn more about an object or attribute by using `explain(obj/attribute, ...)`'
    try:
        l = _autosuggest(a, data, relative_to)
        string = '\n'.join([f'{i}. {s}' for i, s in enumerate(l, start=1)])
    except ImportError:
        raise AttributeError(f"`{a}` not understood.{suffix}")
    raise AttributeError(f"`{a}` not understood, did you mean one of:\n{string}{suffix}")


from typing import Dict, Set

from .actions import TraversalPath
from .hierarchy import *
from .tree import BranchHandler


class Handler:
    def __init__(self, data: 'Data'):
        self.data = data
        self.branch_handler = data.branch_handler  # type: BranchHandler

    def begin_with_heterogeneous(self):
        return HeterogeneousHierarchyFrozenQuery(self, self.branch_handler.entry)

    def hierarchy_from_neo4j_identity(self, htype, identity):
        begin = self.branch_handler.begin(htype.__name__)
        new = begin.add_data(identity)
        identifier_var = new.current_variables[0]
        branch = new.filter('id({h}) = {identifier}', h=begin.current_hierarchy, identifier=identifier_var)
        return DefiniteHierarchyFrozenQuery(self, branch, htype, branch.current_hierarchy, [], None)

    def paths2factor(self, factor_name: str,  plural: bool,
                     start: Type[Hierarchy] = None) -> Tuple[Dict[Type[Hierarchy], Set[TraversalPath]], Type[Hierarchy], bool, str]:
        """
        returns a dictionary of hierarchy: [path,...] and a shared hierarchy
        """
        factor_name = self.data.singular_name(factor_name)
        if start is None:
            starts = set(self.data.factor_hierarchies[factor_name])
            if len(starts) > 1:
                raise AmbiguousPathError(f"{factor_name} could refer to any of {starts}. Please traverse to the parent object first.")
            start = starts.pop()
        pathsetdict, base, factor_name = self.data.find_factor_paths(start, factor_name, plural)
        is_product = factor_name in base.products.keys()
        if len(pathsetdict) > 1:
            names = [i.__name__.lower() for i in pathsetdict.keys()]
            plural = self.data.plural_name(factor_name)
            warn(f"This query for {factor_name} is ambiguous and will return all {plural} "
                 f"for each of the following:\n {', '.join(names)}.\n"
                 f"If this is not what you intended, you can use `.{names[0]}` before `{factor_name}` to choose only {plural} from {names[0]}",
                 SyntaxWarning)
        return pathsetdict, base, is_product, factor_name


    def paths2hierarchy(self, hierarchy_name, plural,
                        start: Type[Hierarchy] = None) -> Tuple[List[TraversalPath], List[Type[Hierarchy]], Type[Hierarchy], Type[Hierarchy]]:
        """
        Returns:
            list of possible paths
            list of hierarchies those paths end with
            the shared start hierarchy
            the shared end hierarchy
        """
        if start is None:
            end = self.data.singular_hierarchies[self.data.singular_name(hierarchy_name)]
            return [], [end], None, end
        return self.data.find_hierarchy_paths(start, self.data.singular_hierarchies[self.data.singular_name(hierarchy_name)], plural)



"""
Queries are a little different
Each time you iterate along, you dont add to the graph straight away.
This gives the Query time to modify the meaning of statements:
without waiting,
    ob.runs[1234] would be [traverse(ob), traverse(run), CopyAndFilter(id=1234)]
with waiting,
    ob.runs[1234] would be [traverse(ob), FilteredMatch(run, id=1234, optional=True)]
    much better!
it treats two chained expressions as one action
"""
import collections
import numpy as np
from copy import copy
from typing import List, TYPE_CHECKING, Union, Tuple, Dict, Any, Iterable

from networkx import NetworkXNoPath

from .align import AlignedQuery
from .base import BaseQuery
from .exceptions import CardinalityError, AttributeNameError, UserError
from .parser import QueryGraph, ParserError
from .utilities import is_regex, dtype_conversion, mask_infs
from .helpers import attributes, objects, find, explain
from astropy.table import Table


if TYPE_CHECKING:
    from weaveio.data import Data


class GenericObjectQuery(BaseQuery):
    def _ipython_key_completions_(self):
        return [i.strip('"') for i in self.__dir__()]


def process_names(given_names, attrs: List['AttributeQuery']) -> List[str]:
    for i, (given_name, attr) in enumerate(zip(given_names, attrs)):
        if given_name is None:
            given_names[i] = attr._factor_name
    if len(set(given_names)) == len(given_names):
        return given_names
    names = [a._factor_name for a in attrs]
    if len(set(names)) == len(names):
        return names
    for i, n in enumerate(names):
        if names.count(n) > 1:
            j = names[:i].count(n)
            names[i] = f"{n}{j}"
    return [a._factor_name for a in attrs]  # todo: this could be better


class ObjectQuery(GenericObjectQuery):
    def __dir__(self) -> List[str]:
        """For autocomplete, only return neighbours and directly owned things"""
        attrs = attributes(self, directly_owned_only=True)
        sing_objs = objects(self, plural=False)
        plur_objs = list(set(objects(self, plural=True)) - set(sing_objs))
        things = sorted(attrs) + sorted(sing_objs) + sorted(plur_objs)
        things = [i if '[' not in i else f'"{i}"' for i in things]
        return [i.lower() for i in things]

    def _get_default_attr(self):
        Obj = self._data.class_hierarchies[self._obj]
        if Obj.idname is not None:
            attr = Obj.idname
        elif len(Obj.products_and_factors) == 1:
            attr = Obj.products_and_factors[0]  # take the only data in it
        elif 'value' in Obj.products_and_factors:
            attr = 'value'
        else:
            raise SyntaxError(f"{self._obj} cannot be returned/identified since it doesn't define any unique idname. "
                              f"If you want to return all singular data for {self._obj} use ...['*']")
        return self.__getitem__(attr)

    def _precompile(self) -> 'TableQuery':
        """
        If the object contains only one factor/product and defines no parents/children, return that
        Otherwise, try to return just the id or error
        """
        return self._get_default_attr()._precompile()

    def _get_all_factors_table(self):
        """
        For use with ['*']
        """
        # single_objs = self._data.all_single_links_to_hierarchy(self._data.class_hierarchies[self._obj])
        # factors = {f if o.__name__ == self._obj else f"{o.singular_name}.{f}" for o in single_objs for f in o.products_and_factors}
        # try:
        #     factors.remove('id')
        #     factors = ['id'] + sorted(factors)
        # except KeyError:
        #     factors = sorted(factors)

        return self._data.class_hierarchies[self._obj].products_and_factors

    def _select_all_attrs(self):
        h = self._data.class_hierarchies[self._data.class_name(self._obj)]
        return self.__getitem__(h.products_and_factors)

    def _traverse_to_generic_object(self):
        """
        obj.generic_obj['specific_type_name']
        filter, shrinks, destructive
        filters based on the type of object to make it more specific
        only shared factors can be accessed
        e.g. `obj.spectra` could be l1singlespectra, l1stackedspectra, l2modelspectra etc
             but `obj.spectra['l1']` ensures that it is only `l1` and only `l1` factors can be accessed
                 `obj.spectra['single']`
        """
        raise NotImplementedError

    def _traverse_to_specific_object(self, obj, want_single):
        """
        obj.obj
        traversal, expands
        e.g. `ob.runs`  reads "for each ob, get its runs"
        """
        plural = self._data.plural_name(obj)
        err_msg = f"There is no singular {obj} relative to {self._obj} try using its plural {plural}"
        try:
            paths, singles = self._get_path_to_object(obj, want_single)
            single = len(singles) == 1 and singles[0]
        except NetworkXNoPath:
            if not want_single:
                err_msg = f"There is no {obj} relative to {self._obj}"
            raise CardinalityError(err_msg)
        if not single and want_single:
            raise CardinalityError(err_msg)
        n = self._G.add_traversal(self._node, paths, obj, single)
        name = plural if want_single else self._data.singular_name(obj)
        return ObjectQuery._spawn(self, n, obj, single=want_single, names=[name])

    def _traverse_by_object_index(self, obj, index):
        """
        obj['obj_id']
        filter, id, shrinks, destructive
        this is a filter which always returns one object
        if the object with this id is not in the hierarchy, then null is returned
        e.g. `obs[1234]` filters to the single ob with obid=1234
        """
        param = self._G.add_parameter(index)
        path, single = self._get_path_to_object(obj, False)
        travel = self._G.add_traversal(self._node, path, obj, single)
        i = self._G.add_getitem(travel, 'id')
        eq, _ = self._G.add_scalar_operation(i, f'{{0}} = {param}', f'id={index}', parameters=[param])
        n = self._G.add_filter(travel, eq, direct=True)
        name = self._data.plural_name(obj) if single else self._data.singular_name(obj)
        return ObjectQuery._spawn(self, n, obj, single=True, names=[name])

    def _traverse_by_object_indexes(self, obj, indexes: List):
        path, single = self._get_path_to_object(obj, False)
        param = self._G.add_parameter(indexes)
        one_id = self._G.add_unwind_parameter(self._node, param, parameters=[param])
        travel = self._G.add_traversal(self._node, path, obj, single, one_id)
        i = self._G.add_getitem(travel, 'id')
        eq, _ = self._G.add_combining_operation('{0} = {1}', 'ids', i, one_id, wrt=travel)
        n = self._G.add_filter(travel, eq, direct=True)
        name = self._data.plural_name(obj) if single else self._data.singular_name(obj)
        return ObjectQuery._spawn(self, n, obj, single=True, names=[name])

    def _select_product(self, attr, want_single):
        attr = self._data.singular_name(attr)
        n = self._G.add_getproduct(self._node, attr)
        return ProductAttributeQuery._spawn(self, n, single=want_single, factor_name=attr, is_products=[True])

    def _select_attribute(self, attr, want_single):
        """
        # obj['factor'], obj.factor
        fetches an attribute from the object and cuts off any other functions after that
        if the factor is not found in the object then the nearest corresponding object is used.
        e.g. `exposure.mjd` returns the mjd held by exposure
             `run.mjd` returns the mjd by a run's exposure (still only one mjd per run though)
             `run.cnames` returns the cname of each target in a run (this is a list per run)
        """
        if self._data.is_product(attr, self._obj):
            return self._select_product(attr, want_single)
        attr = self._data.singular_name(attr)
        n = self._G.add_getitem(self._node, attr)
        return AttributeQuery._spawn(self, n, single=want_single, factor_name=attr)

    def _select_or_traverse_to_attribute(self, attr):
        obj, obj_is_single, attr_is_single = self._get_object_of(attr)  # if factor
        if obj == self._obj:
            return self._select_attribute(attr, attr_is_single)
        r = self._traverse_to_specific_object(obj, obj_is_single)._select_attribute(attr, attr_is_single)
        r._index_node = self._node
        return r

    def _make_table(self, items, all_from_same=False):
        """
        obj['factor_string', AttributeQuery]
        obj['factora', 'factorb']
        obj['factora', obj.obj.factorb]
        obj['factora', 'obj.factorb']
        """
        attrs = []
        for item in items:
            if isinstance(item, dict):
                if len(item) != 1:
                    raise ValueError(f"Must have one key-value pair per item, got {item}")
                name, item = copy(item).popitem()
            if isinstance(item, ObjectQuery):
                item = item._get_default_attr()
            if isinstance(item, AttributeQuery):
                attrs.append(item)
            else:
                new = self.__getitem__(item)
                if isinstance(new, ObjectQuery):
                    new = new._get_default_attr()
                if isinstance(new, list):
                    for a in new:
                        attrs.append(a)
                attrs.append(new)
        force_plurals = [not a._single for a in attrs]
        is_products = [a._is_products[0] for a in attrs]
        n, collected = self._G.add_results_table(self._node, [a._node for a in attrs], force_plurals, treat_equal=all_from_same)
        names = process_names([i if isinstance(i, str) else list(i.keys())[0] if isinstance(i, dict) else None for i in items], attrs)
        t = TableQuery._spawn(self, n, names=names, is_products=is_products, attr_queries=attrs)
        t.one_row = self._G.is_singular_branch_relative_to_splits(n)
        return t

    def _traverse_to_relative_object(self, obj, index, want_singular):
        """
        obj.relative_path
        traversal, expands
        traverses by a specifically labelled relation instead of a name of an object
        e.g. l1stackedspectra.adjunct_spectrum will get the spectrum in the other arm
        CYPHER: (from:From)-[r]->(to:To) WHERE r.id = ...
        the last relationship will have a variable name:
            (Exposure)-->()-[r]->(Run)
            (Run)<-[r]-()<--(Exposure)
        """
        obj, obj_singular = self._normalise_object(obj)
        if obj_singular != want_singular:
            if want_singular:
                raise CardinalityError(f"{obj} is not singular relative to {self._obj}")
        singular_name = self._data.singular_name(index)
        paths, singles = self._get_path_to_object(obj, want_singular)
        n = self._G.add_traversal(self._node, paths, obj, obj_singular)
        relation_id = self._G.add_getitem(n, 'relation_id', 1)
        name = self._G.add_parameter(singular_name)
        eq, _ = self._G.add_scalar_operation(relation_id, f'{{0}} = {name}', 'rel_id', parameters=[name])
        f = self._G.add_filter(n, eq, direct=True)
        return ObjectQuery._spawn(self, f, obj, single=want_singular, names=[index])

    def _filter_by_relative_index(self):
        """
        obj1.obj2s['relative_id']
        filter, relative id, shrinks, destructive
        filters based on the relations between many `obj2`s and the `obj1`
        e.g. `ob.plate_exposures.fibre_exposures.single_spectra['red']` returns the red spectra
        the relative_id are explicitly stated in the database structure
        """
        raise NotImplementedError

    def _getitems(self, items, by_getitem, all_from_same=False):
        # can be called with items signifying columns or where items are indexes, need to distinguish here
        # standardise item list
        if isinstance(items, dict):
            items = [{k: v} for k, v in items.items()]
        else:
            items = list(iter(items))
        items = [[{k: v} for k, v in item.items()] if isinstance(item, dict) else [item] for item in items]
        items = [i for item in items for i in item]
        _items = []
        for item in items:
            if isinstance(item, TableQuery):
                for v in item._attr_queries:
                    _items.append(v)
            elif isinstance(item, dict):
                for k, v in item.items():
                    if isinstance(v, TableQuery):
                        for n, v in zip(v._names, v._attr_queries):  # use table colnames
                            name = n or getattr(v, '_factor_name', '') or ''
                            if k.endswith('_'):
                                _items.append({k+name: v})
                            else:
                                _items.append({name+k: v})
                    else:
                        _items.append({k: v})
            else:
                _items.append(item)
        items = _items
        values = [list(i.values())[0] if isinstance(i, dict) else i for i in items]
        # names = [list(i.keys())[0] if isinstance(i, dict) else i for i in items]
        if not all(isinstance(i, (str, float, int, AttributeQuery, TableQuery, AlignedQuery)) for i in values):
            raise TypeError(f"Cannot index by non str/float/int/AttributeQuery values")
        if all(self._data.is_valid_name(i) or isinstance(i, (AttributeQuery, TableQuery, AlignedQuery)) for i in values):
            return self._make_table(items, all_from_same=all_from_same)
        if any(self._data.is_valid_name(i) for i in values):
            raise SyntaxError(f"You may not mix filtering by id and building a table with attributes")
        # go back and be better
        return self._previous._traverse_by_object_indexes(self._obj, items)

    def _getitem(self, item, by_getitem):
        """
        item can be an id, a factor name, a list of those, a slice, or a boolean_mask
        """
        if isinstance(item, collections.Iterable) and not isinstance(item, (str, BaseQuery)):
            return self._getitems(item, by_getitem)
        elif isinstance(item, slice):
            return self._slice(item)
        elif isinstance(item, AttributeQuery):
            return self._filter_by_mask(item)  # a boolean_mask
        elif not isinstance(item, str):
            return self._previous._traverse_by_object_index(self._obj, item)
        elif item == '*':
            all_factors = self._get_all_factors_table()
            return self._getitems(all_factors, by_getitem, all_from_same=True)
        elif item == '**':
            all_factors = self._data.class_hierarchies[self._obj].products_and_factors
            return self._getitems(all_factors, by_getitem)
        else:
            try:
                return self._select_or_traverse_to_attribute(item)
            except (KeyError, ValueError):
                # if '.' in item and not (item.endswith('.fit') or item.endswith('.fits')):  # split the parts and parse
                #     try:
                #         obj, attr = item.split('.')
                #         return self.__getitem__(obj).__getitem__(attr)
                #     except ValueError:
                #         raise AttributeNameError(f"{item} cannot be parsed as an `obj.attribute`.")
                try: # try assuming its an object
                    obj, single = self._normalise_object(item)
                    if obj == self._obj:
                        return self
                    return self._traverse_to_specific_object(obj, single)
                except (KeyError, ValueError):  # assume its an index of some kind
                    singular = self._data.singular_name(item)
                    if singular in self._data.relative_names:  # if it's a relative id
                        # l1singlespectrum.adjunct gets the adjunct spectrum of a spectrum
                        try:
                            relation = self._data.relative_names[singular][self._obj]
                        except KeyError as e:
                            d = self._data.relative_names[singular]
                            if len(d) > 1:
                                k, v = copy(d).popitem()
                                raise AttributeNameError(f"`{item}` is ambiguous. Choose a specific attribute like `.{self._data.plural_name(k)}.{item}`.")
                            elif not d:
                                raise AttributeNameError(f"`{self._obj}` has no relative relation called `{singular}`")
                            k, relation = copy(d).popitem()
                            return self._traverse_to_specific_object(k, singular == item).__getattr__(item)
                        return self._traverse_to_relative_object(relation.node.__name__, item, singular == item)
                    elif by_getitem and getattr(self._data.class_hierarchies.get(item, None), 'indexable', False):  # otherwise treat as an index
                        return self._previous._traverse_by_object_index(self._obj, item)
                    else:
                        raise AttributeNameError(f"Unknown attribute `{item}`")

    def _getitem_handled(self, item, by_getitem):
        try:
            return self._getitem(item, by_getitem)
        except UserError as e:
            if isinstance(item, AttributeQuery):
                raise e
            self._data.autosuggest(item, self._obj, e)

    def _get_neo4j_id(self):
        n, wrt = self._G.add_scalar_operation(self._node, 'id({0})', 'id')
        return AttributeQuery._spawn(self, n, index_node=wrt, single=True, dtype='int')

    def __getattr__(self, item):
        if isinstance(item, str):
            if item.startswith('_'):
                raise super(ObjectQuery, self).__getattribute__(item)
        return self._getitem_handled(item, False)

    def __getitem__(self, item):
        if isinstance(item, str):
            if item.startswith('_'):
                raise KeyError
        return self._getitem_handled(item, True)

    def __eq__(self, other):
        obj = self._data.class_hierarchies[self._obj]
        potential_id = obj.products_and_factors
        if 'id' in potential_id:
            i = 'id'
        elif 'name' in potential_id:
            i = 'name'
        elif 'value' in potential_id:
            i = 'value'
        else:
            raise AttributeNameError(f"`A {self._obj}` has no attribute `id`, `name`, or `value` to compare with. Choose a specific attribute like `{obj.singular_name}.id`.")
        return self._select_attribute(i, True).__eq__(other)


class TableVariableQuery(ObjectQuery):
    def __dir__(self) -> List[str]:
        return self._table.colnames

    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None,
                 obj: str = None, start: 'Query' = None, index_node=None, single=False, names=None, is_products=None, attrs=None,
                 dtype=None, table=None, *args, **kwargs) -> None:
        super().__init__(data, G, node, previous, obj, start, index_node, single, names, is_products, attrs, dtype, *args, **kwargs)
        self._table = table

    def _precompile(self) -> 'TableQuery':
        return self._get_all_factors_table()

    def _get_all_factors_table(self) -> 'TableQuery':
        return self._make_table(self._table.colnames)

    def _select_all_attrs(self) -> 'TableQuery':
        return self._get_all_factors_table()

    def _traverse_to_generic_object(self):
        raise NotImplementedError

    def _traverse_to_specific_object(self, obj, want_single):
        raise NotImplementedError

    def _traverse_by_object_index(self, obj, index):
        raise NotImplementedError

    def _traverse_by_object_indexes(self, obj, indexes: List):
        raise NotImplementedError

    def _select_product(self, attr, want_single):
        raise NotImplementedError

    def _select_attribute(self, column_name, want_single):
        if column_name not in self._table.colnames:
            raise KeyError(f"`{column_name}` is not a column in the table `{self}`")
        n = self._G.add_getitem(self._node, column_name)
        return AttributeQuery._spawn(self, n, single=want_single, factor_name=column_name)

    def _select_or_traverse_to_attribute(self, attr):
        return self._select_attribute(attr, True)

    def _traverse_to_relative_object(self, obj, index):
        raise NotImplementedError

    def _filter_by_relative_index(self):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError


class Query(GenericObjectQuery):
    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: 'BaseQuery' = None, obj: str = None, start=None) -> None:
        super().__init__(data, G, node, previous, obj, start, 'start')

    def _debug_output(self, skip=0, limit=None, distinct=False, no_cache=False, graph_export_fname=None):
        return super(Query, self)._debug_output(skip, limit, distinct, no_cache, graph_export_fname, exclusive=False)

    def _compile(self):
        raise NotImplementedError(f"{self.__class__} is not compilable")

    def _traverse_to_specific_object(self, obj):
        obj, single = self._normalise_object(obj)
        if single:
            raise CardinalityError(f"Cannot start query with a single object `{obj}`")
        n = self._G.add_start_node(obj)
        return ObjectQuery._spawn(self, n, obj, single=False)

    def _traverse_by_object_index(self, obj, index):
        obj, single = self._normalise_object(obj)
        name = self._G.add_parameter(index)
        if self._node == 0:
            travel = self._G.add_start_node(obj)
        else:
            travel = self._G.add_traversal(self._node, [], obj, single)
        i = self._G.add_getitem(travel, 'id')
        eq, _ = self._G.add_scalar_operation(i, f'{{0}} = {name}', f'id={index}', parameters=[name])
        n = self._G.add_filter(travel, eq, direct=True)
        return ObjectQuery._spawn(self, n, obj, single=True)

    def _traverse_by_object_indexes(self, obj, indexes: List):
        param = self._G.add_parameter(indexes)
        one_id = self._G.add_unwind_parameter(self._node, param, parameters=[param])
        if self._node == 0:
            travel = self._G.add_start_node(obj, one_id)
        else:
            travel = self._G.add_traversal(self._node, [], obj, False, one_id)
        i = self._G.add_getitem(travel, self._data.class_hierarchies[obj].idname)
        eq, _ = self._G.add_combining_operation('{0} = {1}', 'ids', i, one_id)
        n = self._G.add_filter(travel, eq, direct=True)
        return ObjectQuery._spawn(self, n, obj, single=True)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __getattr__(self, item):
        if item in ['keys', 'items', 'values']:
            # catch fastapi jsonable_encoder requests for serialization
            raise AttributeError(f"{self} has no attribute {item}")
        if item.startswith('_'):
            # query attributes are not allowed to start with '_'
            # if we are here, then that means __getattribute__ has failed and we should fail here
            # like a normal object
            raise AttributeError(f"{self} has no attribute {item}")
        try:
            if self._data.is_singular_name(item):
                raise CardinalityError(f"Cannot start a query with a single object `{item}`")
            obj = self._get_object_of(item)[0]
            obj = self._data.plural_name(obj)
            return self._traverse_to_specific_object(obj)._select_attribute(item, True)
        except (KeyError, ValueError):
            try:
                return self._traverse_to_specific_object(item)
            except KeyError as e:
                singular = self._data.singular_name(item)
                d = self._data.relative_names[singular]
                if len(d) > 1:
                    k, v = copy(d).popitem()
                    raise AttributeNameError(f"`{item}` is ambiguous. Choose a specific attribute like `.{self._data.plural_name(k)}.{item}`.")
                elif not d:
                    raise e
                k, relation = copy(d).popitem()
                k = self._data.plural_name(k)
                return self._traverse_to_specific_object(k)._traverse_to_relative_object(relation.node.__name__, singular, singular == item)


class AttributeQuery(BaseQuery):
    one_column = True

    def __repr__(self):
        return f'<{self.__class__.__name__}({self._obj}.{self._factor_name})>'

    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None,
                 obj: str = None, start: Query = None, index_node=None,
                 single=False, factor_name: str = None, dtype: str =None, *args, **kwargs) -> None:
        super().__init__(data, G, node, previous, obj, start, index_node, single, [factor_name], *args, **kwargs)
        self._factor_name = factor_name
        self.dtype = dtype

    def __getitem__(self, item):
        if not isinstance(item, AttributeQuery):
            raise TypeError(f"You can only filter an attribute by another attribute i.e. `lines[lines > 0]` not `lines[0]`")
        op_string = 'CASE WHEN toBoolean(toInteger({1})) THEN {0} ELSE null END'
        try:
            n, wrt = self._G.add_combining_operation(op_string, 'op-filter', self._node, item._node)
        except ParserError:
            raise SyntaxError(f"You may not perform an operation on {self} and {item} since one is not an ancestor of the other")
        return AttributeQuery._spawn(self, n, index_node=n, single=True, dtype=self.dtype)

    def _perform_arithmetic(self, op_string, op_name, *others, expected_dtype=None, returns_dtype=None, parameters=None):
        """
        arithmetics
        [+, -, /, *, >, <, ==, !=, <=, >=]
        performs some maths on factors that have a shared parentage
        e.g. ob1.runs.snrs + ob2.runs.snrs` is not allowed, even if the number of runs/spectra is the same
        e.g. ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red'].snr + ob.l1stackedspectra[ob.l1stackedspectra.camera == 'blue'].snr is not allowed
        e.g. ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red'].snr + ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red'].adjunct.snr is not allowed
        e.g. sum(ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red'].snr, wrt=None) > ob.l1stackedspectra.snr is allowed since one is scalar,
             we take ob.l1stackedspectra as the hierarchy level in order to continue
        e.g. sum(ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red'].snr, wrt=ob) > ob.l1stackedspectra.snr is allowed since there is a shared parent,
             we take ob.l1stackedspectra as the hierarchy level in order to continue
        """
        if any(isinstance(o, ObjectQuery) for o in others):
            raise TypeError(f"Cannot do arithmetic directly on objects")
        if expected_dtype is not None:
            var_templates = [f'{{{i}}}' for i in range(len(others)+1)]
            op_string = dtype_conversion(self.dtype, expected_dtype, op_string, *var_templates)
        masks = [mask_infs(f'{{{iother}}}') if isinstance(other, BaseQuery) else "{iother}" for iother, other in enumerate((self, *others))]
        op_string = op_string.format(*masks)
        if not any(isinstance(o, BaseQuery) for o in others):
            n, wrt = self._G.add_scalar_operation(self._node, op_string, op_name, parameters=parameters)
        else:
            try:
                n, wrt = self._G.add_combining_operation(op_string, op_name, self._node,
                                                         *[other._node for other in others if isinstance(other, BaseQuery)],
                                                         parameters=parameters)
            except ParserError:
                raise SyntaxError(f"You may not perform an operation on {self} and others since one is not an ancestor of the other")
        return AttributeQuery._spawn(self, n, index_node=wrt, single=True, dtype=returns_dtype, factor_name=op_name)

    def _basic_scalar_function(self, name, expected_dtype=None, returns_dtype=None):
        return self._perform_arithmetic(f'{name}({{0}})', name, expected_dtype=expected_dtype, returns_dtype=returns_dtype)

    def _basic_math_operator(self, operator, other, switch=False, out_dtype='number', expected_dtype='number'):
        if not isinstance(other, AttributeQuery):
            other = self._G.add_parameter(other, 'param')
            string_op = f'{other} {operator} {{0}}' if switch else f'{{0}} {operator} {other}'
            ps = [other]
        else:
            string_op = f'{{1}} {operator} {{0}}' if switch else f'{{0}} {operator} {{1}}'
            ps = []
        return self._perform_arithmetic(string_op, operator, other, expected_dtype=expected_dtype, returns_dtype=out_dtype, parameters=ps)

    def _bitwise_operator(self, operator, other, switch=False):
        dtypes = {self.dtype, other.dtype if isinstance(other, AttributeQuery) else type(other)}
        if not dtypes.issubset({'integer', int}):
            raise TypeError(f"For a bitwise operation both sides must be integers")
        if not isinstance(other, AttributeQuery):
            other = self._G.add_parameter(other, 'param')
            string_op = f'apoc.bitwise.op({other}, "{operator}", {{0}})' if switch else f'apoc.bitwise.op({{0}}, "{operator}", {other})'
            ps = [other]
        else:
            string_op = f'apoc.bitwise.op({{1}}, "{operator}", {{0}})' if switch else f'apoc.bitwise.op({{0}}, "{operator}", {{1}})'
            ps = []
        return self._perform_arithmetic(string_op, operator, other, returns_dtype='integer', parameters=ps)

    def _logic_operator(self, logic_operator, bitwise_symbol, other, switch=False):
        """
        If both sides are integers perform bitwise operations
        Otherwise convert to boolean and do normal logical operations
        """
        try:
            # do bitwise
            return self._bitwise_operator(bitwise_symbol, other, switch)
        except TypeError:
            return self._basic_math_operator(logic_operator, other, expected_dtype='boolean', out_dtype='boolean')

    def __and__(self, other):
        return self._logic_operator('and', '&', other)

    def __rand__(self, other):
        return self._logic_operator('and', '&', other, switch=True)

    def __or__(self, other):
        return self._logic_operator('or', '|', other)

    def __ror__(self, other):
        return self._logic_operator('or', '|', other, switch=True)

    def __xor__(self, other):
        return self._logic_operator('xor', '^', other)

    def __rxor__(self, other):
        return self._logic_operator('xor', '^', other, switch=True)

    def __invert__(self):
        return self._basic_scalar_function('not', expected_dtype='boolean', returns_dtype='boolean')

    def __add__(self, other):
        return self._basic_math_operator('+', other)

    def __radd__(self, other):
        return self._basic_math_operator('+', other, switch=True)

    def __mul__(self, other):
        return self._basic_math_operator('*', other)

    def __rmul__(self, other):
        return self._basic_math_operator('*', other, switch=True)

    def __sub__(self, other):
        return self._basic_math_operator('-', other)

    def __rsub__(self, other):
        return self._basic_math_operator('-', other, switch=True)

    def __truediv__(self, other):
        return self._basic_math_operator('/', other, expected_dtype='float', out_dtype='float')

    def __rtruediv__(self, other):
        return self._basic_math_operator('/', other, expected_dtype='float', out_dtype='float', switch=True)

    def __eq__(self, other):
        op = '='
        if isinstance(other, str):
            if is_regex(other):
                op = '=~'
                other = other.strip('/')
        return self._basic_math_operator(op, other, out_dtype='boolean')

    def __ne__(self, other):
        return self._basic_math_operator('<>', other, out_dtype='boolean')

    def __lt__(self, other):
        return self._basic_math_operator('<', other, out_dtype='boolean')

    def __le__(self, other):
        return self._basic_math_operator('<=', other, out_dtype='boolean')

    def __gt__(self, other):
        return self._basic_math_operator('>', other, out_dtype='boolean')

    def __ge__(self, other):
        return self._basic_math_operator('>=', other, out_dtype='boolean')

    def __ceil__(self):
        return self._basic_scalar_function('ceil')

    def __floor__(self):
        return self._basic_scalar_function('floor')

    def __round__(self, ndigits: int):
        return self._perform_arithmetic(f'round({{0}}, {ndigits})', f'round{ndigits}')

    def __neg__(self):
        return self._perform_arithmetic('-{{0}}', 'neg')

    def __abs__(self):
        return self._basic_scalar_function('abs')

    def __iadd__(self, other):
        raise TypeError

    def _precompile(self) -> 'TableQuery':
        if self._index_node == 'start':
            index = self._G.start
        else:
            index = self._index_node
        if index == self._node:  # this happens if we've filtered an attribute by something
            dropna = index
        else:
            dropna = None  # figure it out automatically
        r, collected = self._G.add_results_table(index, [self._node], [not self._single], dropna=dropna)
        a = AttributeQuery._spawn(self, r, self._obj, index, self._single, is_products=self._is_products,
                                  factor_name=self._factor_name)
        if index == 0:
            a.one_row = True
        else:
            a.one_row = self._G.is_singular_branch_relative_to_splits(r)
        return a


class ProductAttributeQuery(AttributeQuery):

    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None,
                 obj: str = None, start: Query = None, index_node=None, single=False, factor_name: str = None, *args, **kwargs) -> None:
        super().__init__(data, G, node, previous, obj, start, index_node, single, factor_name, is_product=[True], *args, **kwargs)

    def _perform_arithmetic(self, op_string, op_name, other=None, expected_dtype=None):
        raise TypeError(f"Binary data products cannot be operated upon. "
                        f"This is because they are not stored in the database")


class TableQuery(BaseQuery):
    def __init__(self, data: 'Data', G: QueryGraph = None, node=None,
                 previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None, obj: str = None,
                 start: Query = None, index_node=None,
                 single=False, attr_queries=None, names=None, *args, **kwargs) -> None:
        super().__init__(data, G, node, previous, obj, start, index_node, single, names, *args, **kwargs)
        self._attr_queries = attr_queries
        self._lookup = {k: v for k, v in zip(self._names, self._attr_queries)}

    def _aggregate(self, wrt, string_op, op_name, predicate=False, expected_dtype=None, returns_dtype=None, remove_infs=None, distinct=False, nulls=False):
        return self._previous._aggregate(wrt, string_op, op_name, predicate, expected_dtype, returns_dtype, remove_infs, distinct, nulls)

    def __getitem__(self, item):
        return self._lookup[item]

    def __add__(self, other: 'TableQuery'):
        if not isinstance(other, TableQuery):
            raise TypeError(f"Can only append columns of another TableQuery to a TableQuery. {other} is a {type(other)}")
        if other._previous is not self._previous:
            raise ValueError(f"Cannot append columns of a table that does not share the same index object.")
        attrs = self._attr_queries + other._attr_queries
        names = self._names + other._names
        duplicate_names = set(self._names) & set(other._names)
        if duplicate_names:
            raise ValueError(f"Cannot append columns of {other} to {self} since they share column names {duplicate_names}. "
                             f"Change their column names with: `table[[{{'new_name': column}}]]` instead of table[[column]]")
        force_plurals = [not a._single for a in attrs]
        is_products = [a._is_products[0] for a in attrs]
        n, collected = self._G.add_results_table(self._previous._node, [a._node for a in attrs], force_plurals)
        t = TableQuery._spawn(self._previous, n, names=names, is_products=is_products, attr_queries=attrs)
        t.one_row = self._G.is_singular_branch_relative_to_splits(n)
        return t


class ListAttributeQuery(AttributeQuery):
    pass



import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'weaveio'
DESCRIPTION = 'WEAVE-IO query interface for accessing WEAVE data'
URL = 'https://github.com/WEAVE-LOFAR/weave-io'
EMAIL = 'shaun.c.read@gmail.com'
AUTHOR = 'philastrophist'

# What packages are required for this module to be executed?
with open('requirements.txt', 'r') as f:
    REQUIRED = f.readlines()

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README.md and use it as the long-description.
# Note: this will only work if 'README.md.rst' is present in your MANIFEST.in file!

with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)

long_description += '\n' + 'machine-readable-change-log\n###########################\n' + about['__version__'] + ': ' + about['__changes__']


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine...')
        os.system('twine upload dist/*')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    #If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    entry_points={'console_scripts': ['weaveio=weaveio.command_line:main']},
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
    setup_requires=['wheel'],
)


import re
from pathlib import Path
from typing import Union, List, Tuple, Dict

from astropy.io import fits
from astropy.io.fits.hdu.base import _BaseHDU

from weaveio.graph import Graph
from weaveio.hierarchy import Hierarchy, Multiple


class File(Hierarchy):
    is_template = True
    idname = 'fname'
    factors = ['path']
    match_pattern = '*.file'
    antimatch_pattern = '^$'
    recommended_batchsize = None
    parts = [None]

    @classmethod
    def length(cls, path, part=None):
        raise NotImplementedError

    def open(self):
        try:
            return fits.open(self.data.rootdir / self.fname)
        except AttributeError:
            return fits.open(self.fname)

    def __init__(self, **kwargs):
        if 'fname' in kwargs:
            kwargs['fname'] = str(kwargs['fname'])
        if 'path' in kwargs:
            kwargs['path'] = str(kwargs['path'])
        else:
            kwargs['path'] = None
        super().__init__(tables=None, **kwargs)

    @classmethod
    def get_batches(cls, path, batch_size, parts: List[Union[str, None]] = None, slc: slice = None):
        if parts is None:
            parts = cls.parts
        parts = sorted({p for p in parts if p in cls.parts})
        if slc is None:
            slc = slice(None, None)
        if batch_size is None:
            return ((slc, part) for part in parts)
        return ((slice(i, i + batch_size), part) for part in parts for i in range(0, cls.length(path, part), batch_size)[slc])

    @classmethod
    def match_file(cls, directory: Union[Path, str], fname: Union[Path, str], graph: Graph):
        """Returns True if the given fname in a given directory can be read by this class of file hierarchy object"""
        fname = Path(fname)
        return re.match(cls.match_pattern, fname.name, re.IGNORECASE) and not re.match(cls.antimatch_pattern, fname.name, re.IGNORECASE)

    @classmethod
    def match_files(cls,  directory: Union[Path, str], graph: Graph):
        """Returns all matching files within a directory"""
        return (f for f in Path(directory).rglob('*.fit*') if cls.match_file(directory, f, graph))

    @classmethod
    def check_mos(cls, path):
        header = fits.open(path)[0].header
        try:
            return 'IFU' not in header['OBSMODE'] and (header.get('OBSTYPE', '') in ['TARGET', ''])
        except KeyError as e:
            raise KeyError(f"File {path} does not contain OBSTYPE keyword") from e

    @classmethod
    def read(cls, directory: Union[Path, str], fname: Union[Path, str], slc: slice = None, part=None) -> 'File':
        raise NotImplementedError

    @classmethod
    def read_hdus(cls, directory: Union[Path, str], fname: Union[Path, str],
                  **hierarchies: Union[Hierarchy, List[Hierarchy]]) -> Tuple[Dict[int,'HDU'], 'File', List[_BaseHDU]]:
        path = Path(directory) / Path(fname)
        relative_path = path.relative_to(Path(directory))
        file = cls(fname=path.name, path=str(relative_path), **hierarchies)
        hdus = [i for i in fits.open(path)]
        if len(hdus) != len(cls.hdus):
            raise TypeError(f"Class {cls} asserts there are {len(cls.hdus)} HDUs ({list(cls.hdus.keys())})"
                            f" whereas {path} has {len(hdus)} ({[i.name for i in hdus]})")
        hduinstances = {}
        for i, ((hduname, hduclass), hdu) in enumerate(zip(cls.hdus.items(), hdus)):
            hduinstances[i] = hduclass.from_hdu(hduname, hdu, i, file)
        return hduinstances, file, hdus

    def read_product(self, product_name):
        self.build_index()
        return getattr(self, f'read_{product_name}')()


class HDU(Hierarchy):
    is_template = True
    parents = [File]
    factors = ['extn', 'name']
    identifier_builder = ['file', 'extn', 'name']
    products = ['header', 'data']

    @classmethod
    def from_hdu(cls, name, hdu, extn, file):
        input_dict = {}
        input_dict[cls.parents[0].singular_name] = file
        input_dict['extn'] = extn
        input_dict['name'] = name
        hdu = cls(**input_dict)
        for product in cls.products:
            hdu.attach_product(product, hdu)
        return hdu


class BaseDataHDU(HDU):
    is_template = True


class PrimaryHDU(HDU):
    products = ['header']


class TableHDU(BaseDataHDU):
    pass


class BinaryHDU(BaseDataHDU):
    pass



from pathlib import Path
from typing import List, Union

from astropy.io import fits
from astropy.table import Table as AstropyTable, Column, Row as AstropyRow
import pandas as pd
import numpy as np
from py2neo.cypher import Cursor, Record
from tqdm import tqdm

from weaveio.basequery.common import FrozenQuery
from weaveio.basequery.dissociated import Dissociated
from weaveio.basequery.tree import Branch
from weaveio.writequery import CypherVariable, CypherQuery


class Row(AstropyRow):
    def __getattr__(self, attr):
        if attr in self.colnames:
            return self[attr]
        return super(Row, self).__getattr__(attr)


class Table(AstropyTable):  # allow using `.` to access columns
    Row = Row

    def __getattr__(self, attr):
        if attr in self.colnames:
            return self[attr]
        return super(Table, self).__getattr__(attr)


def replace_with_data(row, files):
    hdus = files[row['sourcefile']]
    hdu = hdus[int(row['extn'])]
    if not pd.isnull(row['column_name']):
        table = Table(hdu.data)
        table = table[row['column_name']]
    else:
        table = hdu.data
    if not pd.isnull(row['index']):
        return table[int(row['index'])]
    return table


def safe_name(name):
    return '__dot__'.join(name.split('.'))


class FactorFrozenQuery(Dissociated):
    def __init__(self, handler, branch: Branch, factors: List[str], factor_variables: List[CypherVariable],
                 plurals: List[bool], is_products: List[bool], parent: FrozenQuery = None):
        super().__init__(handler, branch, factor_variables[0], parent)
        self.factors = factors
        self.factor_variables = factor_variables
        self.plurals = plurals
        self.is_products = is_products

    def _filter_by_boolean(self, boolean_filter: 'Dissociated'):
        new = self._make_filtered_branch(boolean_filter)
        return self.__class__(self.handler, new, self.factors, self.factor_variables, self.plurals,
                              self.is_products, self)

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            variables = {safe_name(k): v for k, v in zip(self.factors, self.factor_variables)}
            return query.returns(**variables)

    def __repr__(self):
        if isinstance(self.factors, tuple):
            factors = f'{self.factors}'
        else:
            factors = f'[{self.factors}]'
        return f'{self.parent}{factors}'

    def _parse_products(self, df: pd.DataFrame):
        """
        Take a pandas dataframe and replace the structure of ['sourcefile', 'extn', 'index', 'column_name']
        with the actual data
        """
        product_columns = []
        for colname, is_product in zip(df.columns, self.is_products):
            if is_product:
                df[colname] = df[colname].apply(lambda x: [None]*4 if len(x) == 0 else x[0])
                split = df[colname].apply(pd.Series)
                split.columns = ['sourcefile', 'extn', 'index', 'column_name']
                split.index.name = 'rown'
                split['colname'] = colname
                split.set_index('colname', append=True, inplace=True)
                product_columns.append(split)
        if len(product_columns):
            sourcefiles = {}
            concat = pd.concat(product_columns)
            concat = concat.applymap(lambda x: np.nan if x is None else x).dropna(how='all')
            for fname in tqdm(concat.sourcefile.drop_duplicates(), desc='reading fits files'):
                if fname is not None:
                    path = Path(self.handler.data.rootdir) / fname
                    sourcefiles[fname] = fits.open(path)
            data = concat.apply(replace_with_data, files=sourcefiles, axis='columns').sort_index(level=0)  # type: pd.Series
            df.index.name = 'rown'
            for name, group in data.groupby('colname'):
                df[name] = group.droplevel(1)
        return df

    def _post_process(self, result: Union[Cursor, Record], squeeze: bool = True) -> Table:
        if isinstance(result, Record):
            df = pd.DataFrame([dict(result)])
        else:
            df = pd.DataFrame(list(map(dict, result)))
        df = self._parse_products(df)
        # replace lists with arrays
        for c in df.columns:
            if df.dtypes[c] == 'O' and not isinstance(df[c].iloc[0], str):
                try:
                    if np.all(df[c].apply(len) == 0):
                        df[c] = np.nan
                except TypeError:  # smelly but skips missing data rows
                    pass
                df[c] = df[c].apply(np.asarray)
        table = Table.from_pandas(df)
        for colname, plural, is_product in zip(df.columns, self.plurals, self.is_products):
            if plural or is_product or df.dtypes[colname] == 'O':
                shapes = set(map(np.shape, df[colname]))
                if len(shapes) == 1:  # all the same length
                    table[colname] = Column(np.stack(df[colname].values), name=colname, shape=shapes.pop(), length=len(df))
        if squeeze and len(table) == 1:
            table = table[0]
        return table


class SingleFactorFrozenQuery(FactorFrozenQuery):
    def __init__(self, handler, branch: Branch, factor: str, factor_variable: CypherVariable, is_product: bool, parent: FrozenQuery = None):
        super().__init__(handler, branch, [factor], [factor_variable], [False], [is_product], parent)

    def _post_process(self, result: Union[Cursor, Record], squeeze: bool = True):
        table = super()._post_process(result, squeeze=False)
        column = table[table.colnames[0]].data
        if (len(column) == 1 and squeeze) or isinstance(result, Record):
            return column[0]
        return column


class TableFactorFrozenQuery(FactorFrozenQuery):
    """
    A matrix of different factors against different hierarchy instances
    This is only possible if the hierarchies each have only one of the factors
    """
    def __init__(self, handler, branch, factors, factor_variables, plurals, is_products,
                 return_keys: List[str] = None, parent: 'FrozenQuery' = None):
        super().__init__(handler, branch, factors, factor_variables, plurals, is_products, parent)
        self.return_keys = return_keys

    def _prepare_query(self) -> CypherQuery:
        with super()._prepare_query() as query:
            variables = {safe_name(k): v for k, v in zip(self.return_keys, self.factor_variables)}
            return query.returns(**variables)

    def _post_process(self, result: Union[Cursor, Record], squeeze: bool = True) -> Table:
        t = super()._post_process(result, squeeze=False)
        if len(t):
            t = Table(t)
            t.rename_columns(t.colnames, self.return_keys)
        else:
            t = Table(names=self.return_keys)
        if isinstance(result, Record):
            return t[0]
        return t

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        if isinstance(item, str):
            try:
                i = self.factors.index(item)
            except ValueError:
                raise KeyError(f"{item} is not a factor contained within {self}. Only {self.factors} are accessible.")
            else:
                return SingleFactorFrozenQuery(self.handler, self.branch, item, self.factor_variables[i], self.is_products[i], self)
        else:
            return super(TableFactorFrozenQuery, self).__getitem__(item)



from collections import OrderedDict, Counter, defaultdict
from pathlib import Path
from typing import List, Union, Tuple

import numpy as np
import py2neo
from astropy.io import fits, registry
from astropy.table import Table as AstropyTable, Row as AstropyRow, MaskedColumn as AstropyMaskedColumn
from py2neo.cypher import Cursor


def ragged_array(data):
    """
    Given an object array of different length lists/arrays (indexed on the first axis) return a ragged array
    where unaligned elements are filled with NaN.
    e.g.:
    >>> ragged_array([[1, 2, 3], [4, 5, 6, 7]])
    np.mask.array([[1, 2, 3, --], [4, 5, 6, 7]], mask=[[False, False, False, True], [False, False, False, False]])
    """
    darrays = [np.ma.asarray(d) for d in data]
    dtypes = {d.dtype for d in darrays if not np.all(d.mask)}
    if not dtypes:
        dtype = np.float_
    else:
        dtype = np.stack([np.array([], dtype=d) for d in dtypes]).dtype
    shapes = [d.shape for d in darrays]
    if len(set(shapes)) == 1:
        return data
    maxshape = np.max(shapes, axis=0)
    array = np.ma.empty((len(data), *maxshape), dtype=dtype)
    array.mask = True
    slcs = [tuple(slice(0, s) for s in shape_tuple) for shape_tuple in shapes]
    for i, (slc, d) in enumerate(zip(slcs, data)):
        array.__setitem__((i, *slc), d)
    return array


def ragged_column(data, name):
    return MaskedColumn(ragged_array(data), name=name)


class ColnameParser:
    def __init__(self, parent, partial_colname: str):
        self.parent = parent
        self.partial_colname = partial_colname

    def __getattr__(self, colname):
        return self.parent.__getattr__(self.partial_colname + '.' + colname)


class DotHandlerMixin:
    def __getattr__(self, attr):
        if attr in self.colnames:
            return self[attr]
        if any(colname.startswith(attr) for colname in self.colnames if '.' in colname):
            return ColnameParser(self, attr)
        return super(DotHandlerMixin, self).__getattribute__(attr)


class Row(DotHandlerMixin, AstropyRow):
    pass


class MaskedColumn(AstropyMaskedColumn):
    def apply(self, func, *args, **kwargs) -> 'MaskedColumn':
        """
        Apply a function to each row in the column. Returns nd array if possible otherwise an unstructured array.
        :param func: Callable to apply to each row. Takes an array/scalar depending on the type of the column.
        """
        return ragged_column([func(d, *args, **kwargs) for d in self.data], self.name)

    def masked(self, mask, inplace=False):
        if inplace:
            c = self
        else:
            c = self.copy()
        c.mask |= mask
        return c

    def filtered(self, filt, inplace=False):
        if inplace:
            c = self
        else:
            c = self.copy()
        c.mask |= ~filt
        return c


class Table(DotHandlerMixin, AstropyTable):  # allow using `.` to access columns
    Row = Row
    Column = MaskedColumn

    def apply(self, func, *args, **kwargs) -> 'MaskedColumn':
        """
        Apply a function to each row in the column. Returns nd array if possible otherwise an unstructured array.
        :param func: Callable to apply to each row. Takes an array/scalar depending on the type of the column.
        """
        return ragged_column([func(row, *args, **kwargs) for row in self], func.__name__)

class ArrayHolder:
    def __init__(self, array):
        self.array = array


def vstack_rows(rows: List[Tuple[List, List[bool], List[str]]], *args, **kwargs) -> Table:
    # for each column, remove null rows, make table, put nulls back in
    columns, names = zip(*rows)
    names = names[0]
    columns = list(zip(*columns))
    duplicate_names = [n for n, i in Counter(names).items() if i > 1]
    counter = defaultdict(int)
    _names = []
    for n in names:
        if n in duplicate_names:
            _names.append(f"{n}{counter[n]}")
            counter[n] += 1
        else:
            _names.append(n)
    return Table([ragged_column(c, f"{n}{counter.get(n, '')}") for c, n in zip(columns, _names)])

def int_or_slice(x: Union[int, float, slice, None]) -> Union[int, slice]:
    if isinstance(x, (int, float)):
        return int(x)
    elif isinstance(x, slice):
        return x
    else:
        return slice(None, None)

class FileHandler:
    def __init__(self, rootdir: Union[Path, str], max_concurrency: int = 1000):
        self.rootdir = Path(rootdir)
        self.max_concurrency = max_concurrency
        self.files = OrderedDict()

    def read(self, filename: Union[Path, str], ext: Union[float, int, str] = None, index: Union[float, int, slice] = None,
             key: Union[str, int, float, slice] = None, header_only=False):
        f = self.open_file(filename)
        if ext is None:
            ext = 0
        ext = ext if isinstance(ext, str) else int(ext)
        hdu = f[ext]
        index = int_or_slice(index)
        key = key if isinstance(key, str) else int_or_slice(key)
        if header_only:
            if key is not None:
                return hdu.header[key]
            return hdu.header
        return hdu.data[index][key]


    def open_file(self, filename: Union[Path, str]):
        filename = self.rootdir / Path(filename)
        if filename in self.files:
            return self.files[filename]
        else:
            if len(self.files) >= self.max_concurrency:
                self.close_file(next(iter(self.files)))
            self.files[filename] = fits.open(str(filename), memmap=True)
            return self.files[filename]

    def close_file(self, filename: Path):
        if filename in self.files:
            del self.files[filename]

    def close_all(self):
        for filename in list(self.files):
            self.close_file(filename)

    def __del__(self):
        self.close_all()


def recursive_replace_None(l: List) -> List:
    if not isinstance(l, list):
        return np.ma.masked if l is None else l
    for i, item in enumerate(l):
        l[i] = recursive_replace_None(item)
    return l


class RowParser(FileHandler):
    def parse_product_row(self, row: py2neo.cypher.Record, names: List[Union[str, None]], is_products: List[bool],
                          as_row: bool):
        """
        Take a pandas dataframe and replace the structure of ['fname', 'extn', 'index', 'key', 'header_only']
        with the actual data
        """
        columns = []
        colnames = []
        for value, cypher_name, name, is_product in zip(row.values(), row.keys(), names, is_products):
            if is_product:
                if value is not None:
                    if isinstance(value[0], list):  # i.e. its a list of product addresses that have been collected
                        value = [self.read(*v) for v in value]
                    else:
                        value = self.read(*value)
            value = recursive_replace_None(value)
            name = cypher_name if name is None or name == 'None' else name
            mask = value is None or np.size(value) == 0
            try:
                mask = mask | ~np.isfinite(value)
            except TypeError:
                pass
            value = np.ma.where(~mask, np.ma.asarray(value), np.ma.masked)
            columns.append(value)
            colnames.append(name)
        if as_row:
            columns = [MaskedColumn([value]) for value in columns]
            return Table(columns, names=colnames)[0]
        return columns, colnames

    def iterate_cursor(self, cursor: Cursor, names: List[Union[str, None]], is_products: List[bool], as_row: bool):
        for row in cursor:
            yield self.parse_product_row(row, names, is_products, as_row)

    def parse_to_table(self, cursor: Cursor, names: List[str], is_products: List[bool]):
        rows = list(self.iterate_cursor(cursor, names, is_products, False))
        if not rows:
            return Table([MaskedColumn([], name=name) for name in names])
        return vstack_rows(rows)


def apply(obj, func, *args, **kwargs):
    """Applies the function to a table or column"""
    return getattr(obj, 'apply')(func, *args, **kwargs)

def filtered(column, filt):
    """Filters a masked column based on a boolean array where False means masked"""
    return column.filtered(filt)

def masked(column, mask):
    """Masks a masked column based on a boolean array where True means masked"""
    return column.masked(mask)


def identifier(origin, filepath, fileobj, *args, **kwargs):
    return registry._identifiers[('fits', AstropyTable)](origin, filepath, fileobj, *args, **kwargs)


def reader(input, hdu=None, astropy_native=False, memmap=False, character_as_bytes=True):
    original = registry.get_reader('fits', AstropyTable)
    table = original(input, hdu=hdu, astropy_native=astropy_native, memmap=memmap, character_as_bytes=character_as_bytes)
    for colname, fill_value, nan_mask in zip(table.meta['_MASKED'], table.meta['_FILL'], table.meta['_NAN']):
        if nan_mask:
            fill_value = np.nan
            mask = np.isnan(table[colname])
        else:
            mask = table[colname] == fill_value
        table[colname] = MaskedColumn(table[colname], fill_value=fill_value, mask=mask)
    return table

def writer(input, output, overwrite=False):
    original = registry.get_writer('fits', AstropyTable)
    input.meta['_masked'] = []
    input.meta['_fill'] = []
    input.meta['_nan'] = []
    for colname in input.colnames:
        if isinstance(input[colname], MaskedColumn):
            input.meta['_masked'].append(colname)
            try:
                if np.isnan(input[colname].fill_value):
                    input.meta['_nan'].append(True)
                    input.meta['_fill'].append(0.)
                    continue
            except TypeError:
                pass
            input.meta['_fill'].append(input[colname].fill_value)
            input.meta['_nan'].append(False)
    return original(input, output, overwrite)


registry.register_identifier('fits', Table, identifier, force=True)
registry.register_reader('fits', Table, reader, force=True)
registry.register_writer('fits', Table, writer, force=True)


import itertools
import logging
import time
import warnings
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from copy import copy
from typing import Tuple, List, Dict, Any, Union, TYPE_CHECKING
from warnings import warn

import networkx as nx
from networkx import NetworkXNoPath, NodeNotFound
from py2neo.cypher import Cursor

from weaveio.hierarchy import Hierarchy

from .exceptions import AmbiguousPathError, CardinalityError, DisjointPathError, AttributeNameError
from .parser import QueryGraph
from .results import Table, AstropyTable
from ..path_finding import collapse_classes_to_superclasses

if TYPE_CHECKING:
    from .objects import ObjectQuery, Query, AttributeQuery
    from ..data import Data


@contextmanager
def logtime(name):
    start = time.perf_counter()
    yield
    logging.info(f"{name} took {time.perf_counter() - start: .3f} seconds")


def get_groupids(graph, params):
    groupids = []
    names = []
    for name, query in graph.groupbys.items():
        if name in params:
            groupids.append(query._iterate(limit=None, skip=0, distinct=True, no_groups=True))
            names.append(name)
    return groupids, names

def _execute_groups(groupids, names, params):
    for values in itertools.product(*groupids):
        for name, value in zip(names, values):
            params[name] = value
        yield values, params


SplitQuery = namedtuple('SplitQuery', ['index', 'query'])
SplitResult = namedtuple('SplitResult', ['index', 'result'])


class QueryFunctionBase:
    def _precompile(self):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}()'




class BaseQuery(QueryFunctionBase):
    one_row = False
    one_column = False

    def __new__(cls, *args, **kwargs):
        from fastapi.encoders import ENCODERS_BY_TYPE
        ENCODERS_BY_TYPE[cls] = lambda x: [x._data, x._node]
        return super().__new__(cls)

    def _debug_output(self, skip=0, limit=None, distinct=False, no_cache=False, graph_export_fname=None,
                      exclusive=True):
        n = self._precompile()
        cached_params = n._get_cached_parameters()
        placeholder_params = n._G.dependency_parameters(self._node)
        params = {**placeholder_params, **cached_params}
        c = n._to_cypher(skip, limit, distinct, no_cache)
        params = n._prepare_parameters(c, params)
        if graph_export_fname is not None:
            if exclusive:
                n._G.export(graph_export_fname, n._node)
            else:
                n._G.export(graph_export_fname)
        return '\n'.join(c), self._G.G.nodes[n._node]['ordering'], params

    def raise_error_with_suggestions(self, obj, exception: Exception):
        self._data.autosuggest(obj, self._obj, exception)
        raise exception

    def __repr__(self):
        try:
            return f'<{self.__class__.__name__}({self._previous._obj}-{self._obj})>'
        except AttributeError:
            return f'<{self.__class__.__name__}({self._obj})>'

    def _precompile(self) -> 'BaseQuery':
        return self

    def _prepare_parameters(self, lines, parameters):
        params = {}
        for k, v in parameters.items():
            if any(k in l for l in lines):
                if isinstance(v, AstropyTable):
                    cols = [c for c in v.colnames if any(c in l for l in lines)]
                    assert len(cols), "No columns are used, but table is used. This should not happen"
                    params[k] = v[cols]
                else:
                    params[k] = v
        return params

    def _to_cypher(self, skip, limit, distinct, no_cache=False) -> List[str]:
        """
        returns the cypher lines, cypher parameters, names of columns, expect_one_row, expect_one_column
        """
        lines = self._G.cypher_lines(self._node, no_cache)
        if distinct:
            lines[-1] = 'RETURN DISTINCT'.join(lines[-1].rsplit('RETURN', 1))
        if skip > 0:
            lines.append(f"\nSKIP {skip}")
        if limit is not None:
            lines.append(f"\nLIMIT {limit}")
        return lines

    def _compile(self, skip, limit, distinct) -> Tuple['BaseQuery', List[str]]:
        with logtime('compiling'):
            r = self._precompile()
            return r, r._to_cypher(skip, limit, distinct)

    def _execute_single_query(self, query, cypher, params):
        return query._data.graph.execute(cypher, **{k.replace('$', ''): v for k, v in params.items()})

    def _get_cached_parameters(self):
        params = {}
        for node in self._G.restricted(self._node).nodes:
            params.update(self._G.G.nodes[node].get('cached_params', {}))
        return params

    def _execute(self, skip=0, limit=None, distinct=False, no_groups=False):
        """
        returns a new query (using _precompile) and an iterator over cursors.
        For queries that have not been `split`, this is a single cursor.
        For queries that have been `split`, there will be multiple cursors.
        :param skip: number of rows to skip
        :param limit: number of rows to limit to
        :param distinct: whether to return distinct rows
        :param no_groups: whether to return a single cursor or a generator of cursors
        """
        with logtime('executing'):
            cached_params = self._get_cached_parameters()
            placeholder_params = self._G.dependency_parameters(self._node)
            params = {**placeholder_params, **cached_params}
            if not no_groups and any(v == '<placeholder>' for v in params.values()):
                groupids, names = get_groupids(self._G, params)
                if groupids:
                    return _execute_groups(groupids, names, params), self
            new, lines = self._compile(skip, limit, distinct)
            cypher = '\n'.join(lines)
            params = self._prepare_parameters(lines, params)
            return self._execute_single_query(new, cypher, params), new

    def _iterate_groups(self, generator, query):
        for split_indexes, params in generator:
            copied = copy(query)  # copy the query, so that the parameters are not shared between the cursors
            query._G.G.nodes[copied._node]['cached_params'] = params  # store the groups for when the query is executed
            yield SplitQuery(split_indexes, copied)

    def _iterate(self, skip=0, limit=None, distinct=False, no_groups=False):
        cursor_or_gen, new = self._execute(skip, limit, distinct, no_groups)  # produces either a cursor or a generator of cursors
        if not isinstance(cursor_or_gen, Cursor):
            yield from self._iterate_groups(cursor_or_gen, new)
        else:
            rows = new._data.rowparser.iterate_cursor(cursor_or_gen, new._names, new._is_products, True)
            i = None
            with logtime('total streaming (by iteration)'):
                for i, row in enumerate(rows):
                    yield new._post_process_row(row)
            if limit == i:
                warnings.simplefilter("always")
                warn(f"This query has been capped to {limit} rows but the result is larger than that. "
                     f"Consider using the `limit` parameter to better limit the result or set `limit` to None to get "
                     f"the full result (although this is not recommended).")

    def __iter__(self):
        yield from self._iterate()

    def _to_table(self, skip=0, limit=None, distinct=False) -> Tuple[Table, 'BaseQuery']:
        with logtime('total streaming'):
            cursor_or_gen, new = self._execute(skip, limit, distinct)  # produces either a cursor or a generator of cursors
            if not isinstance(cursor_or_gen, Cursor):
                return [SplitResult(split_indexes, query._to_table(skip, limit, distinct)) for split_indexes, query in self._iterate_groups(cursor_or_gen, new)], new
            return new._data.rowparser.parse_to_table(cursor_or_gen, new._names, new._is_products), new

    def _post_process_table(self, result, squeeze):
        if isinstance(result, SplitResult):
            return result.index, self._post_process_table(result.result, squeeze)
        if not squeeze:
            return result
        if self.one_column:
            result = result[result.colnames[0]].data
        if self.one_row:
            result = result[0]
        return result

    def _post_process_row(self, row):
        if self.one_column:
            return row[row.colnames[0]]
        return row

    def __call__(self, skip=0, limit=1000, distinct=False, squeeze=True, **kwargs):
        tbl, new = self._to_table(skip, limit, distinct)
        if isinstance(tbl, list):
            return tbl
        tbl = new._post_process_table(tbl, squeeze)
        try:
            if not len(tbl):
                return tbl
        except TypeError:
            return tbl  # not a table, just a value, so return it
        if limit == 1 and squeeze:
            return tbl[0]
        try:
            if limit == len(tbl):
                warnings.simplefilter("always")
                warn(f"This query has been capped to {limit} rows but the result is larger than that. "
                     f"Consider using the `limit` parameter to better limit the result or set `limit` to None to get "
                     f"the full result (although this is not recommended).")
        except TypeError:
            pass
        return tbl


    def __init__(self, data: 'Data', G: QueryGraph = None, node=None, previous: Union['Query', 'AttributeQuery', 'ObjectQuery'] = None,
                 obj: str = None, start: 'Query' = None, index_node = None,
                 single=False, names=None, is_products=None, attrs=None, dtype=None, *args, **kwargs) -> None:
        from .objects import ObjectQuery
        self._single = single
        self._data = data
        if G is None:
            self._G = QueryGraph()
        else:
            self._G = G
        if node is None:
            self._node = self._G.start
        else:
            self._node = node
        self._previous = previous
        self.__cypher = None
        self._index_node = index_node
        self._obj = obj
        if previous is not None:
            if self._index_node is None:
                if isinstance(previous, ObjectQuery):
                    self._index_node = previous._node
                elif self._index_node != 'start':
                    self._index_node = previous._index_node
            if obj is None:
                self._obj = previous._obj
        if start is None:
            self._start = self
        else:
            self._start = start
        if self._obj is not None:
            if not self._obj.startswith('_'):  # system name for other purposes
                self._obj = self._normalise_object(self._obj)[0]
        self._names = [] if names is None else names
        self._is_products = [False]*len(self._names) if is_products is None else is_products
        self.attrs = attrs
        self.dtype = dtype

    def _get_object_of(self, maybe_attribute: str) -> Tuple[str, bool, bool]:
        """
        Given a str that might refer to an attribute, return the object that contains that attribute
        and also whether the attribute and the object are singular.
        plurality table:
            inferred obj | attr
                s        |  s      (i.e. run.obid -> run.ob.obid)
                p        |  p      (i.e. ...-> run.l1specs.snrs) - cannot be inferred just from a name
                s        |  p      (i.e. run.obids -> run.ob.obids)
                p        |  s      (i.e. run.snr -> run.l1spec.snr) - is not allowed
        attribute storage scenarios:
            * attribute is contained in this object - return self
            * attribute is contained in exactly one object - return object
            * attribute is contained in more than one object and ...
                * any of those objects don't have the same inherited class - raise AmbiguousError
                * only subclass objects and all subclass objects from a superclass contain the attribute - return the superclass
                * only subclass objects and not all subclass objects from a superclass contain the attribute - return the superclass
        If a singular path exists, use that
        :return: obj, obj_is_singular, attr_is_singular
        """
        single_name = self._data.singular_name(maybe_attribute)
        if self._obj is not None:
            if self._data.class_hierarchies[self._obj] in self._data.factor_hierarchies[single_name]:
                return self._obj, True, self._data.is_singular_name(maybe_attribute)
        want_single = single_name == maybe_attribute
        # try to get singular paths
        hs = {h for h in self._data.factor_hierarchies[single_name] if self._has_path_to_object(h, True)}
        if len(hs) == 0: # if no singular paths then get all paths
            hs = {h for h in self._data.factor_hierarchies[single_name] if self._has_path_to_object(h, want_single)}
        hs = set(collapse_classes_to_superclasses(self._data.hierarchy_graph, hs))
        if len(hs) == 1:
            obj = hs.pop().__name__
        elif not hs:
            raise AmbiguousPathError(f"There is no path to any object that contains `{single_name}`. {self._data.factor_hierarchies[single_name]} are not reachable")
        else:
            raise AmbiguousPathError(f"`{single_name}` is contained within more than one unrelated object {[h.__name__ for h in hs]}. Be specific.")
        if self._obj is None:
            return obj, True, False
        if not self._data.is_factor_name(maybe_attribute):
            raise ValueError(f"{maybe_attribute} is not a valid attribute name")
        _, obj_is_singulars = self._get_path_to_object(obj, False)
        obj_is_singular = len(obj_is_singulars) == 1 and obj_is_singulars[0]
        attr_is_singular = self._data.is_singular_name(maybe_attribute)
        if obj_is_singular and attr_is_singular:
            pass  # run.obid -> run.ob.obid
        elif not obj_is_singular and not attr_is_singular:
            pass
        elif obj_is_singular and not attr_is_singular:
            pass  # run.obids -> run.ob.obids
        else:
            # ob.runid -> Error
            plural_name = self._data.plural_name(obj)
            original = self._data.singular_name(self._obj)
            raise CardinalityError(f"Requested one `{single_name}` from `{original}` "
                                   f"when `{original}` has several `{plural_name}`")
        return obj, obj_is_singular, attr_is_singular

    def _normalise_object(self, obj: str):
        """
        returns the __name__ of the respective class of the obj and whether it is plural/singular.
        If not a class, raise KeyError
        """
        if obj in self._data.class_hierarchies:
            return obj, True
        singular = self._data.singular_name(obj)
        if singular not in self._data.singular_hierarchies:
            raise KeyError(f"{obj} is not a valid object name")
        try:
            h = self._data.plural_hierarchies[obj.lower()]
            singular = False
        except KeyError:
            h = self._data.singular_hierarchies[obj.lower()]
            singular = True
        return h.__name__, singular

    @property
    def _cypher(self):
        if self.__cypher is None:
            self.__cypher = self._G.cypher_lines(self._node)
        return self.__cypher

    @property
    def _result_graph(self):
        return self._G.restricted(self._node)

    def _plot_graph(self, fname):
        return self._G.export(fname, self._node)

    @classmethod
    def _spawn(cls, parent: 'BaseQuery', node, obj=None, index_node=None, single=False, *args, **kwargs):
        return cls(parent._data, parent._G, node, parent, obj, parent._start, index_node, single, *args, **kwargs)

    def _get_path_to_object(self, obj, want_single) -> Tuple[List[str], List[bool]]:
        return self._data.paths_to_hierarchy(self._obj, obj, want_single)

    def _has_path_to_object(self, obj, want_single) -> bool:
        if self._obj is None:
            return True
        try:
            return len(self._get_path_to_object(obj, want_single)[0]) > 0
        except NetworkXNoPath:
            return False

    def _slice(self, slc):
        """
        obj[slice]
        filter, shrinks, destructive
        filter by using HEAD/TAIL/SKIP/LIMIT
        e.g. obs.runs[:10] will return the first 10 runs for each ob (in whatever order they were in)
        you must use query(skip, limit) to request a specific number of rows in total since this is unrelated to the actual query
        """
        raise NotImplementedError

    def _filter_by_mask(self, mask, single=None, split_node=False):
        """
        obj[boolean_filter]
        filter, shrinks, destructive
        filters based on a list of True/False values constructed beforehand, parentage of the booleans must be derived from the obj
        e.g. `ob.l1stackedspectra[ob.l1stackedspectra.camera == 'red']` gives only the red stacks
             `ob.l1stackedspectra[ob.l1singlespectra == 'red']` is invalid since the lists will not be the same size or have the same parentage
        """
        if mask.dtype != 'boolean':
            raise TypeError(f"The predicate in a filter `query[predicate]`, must be of type boolean. {mask} is of type {mask.dtype}. ")
        single = single if single is not None else self._single
        try:
            n = self._G.add_filter(self._node, mask._node, direct=False, force_single=single, split_node=split_node)
        except SyntaxError:
            raise DisjointPathError(f"{self} cannot be filtered by {mask} since `{self._obj}` != `{mask._obj}`."
                                    f"This may be because a nonsensical `wrt` has specified in an aggregation.")
        return self.__class__._spawn(self, n, single=single)

    def _aggregate(self, wrt, string_op, op_name, predicate=False, expected_dtype=None, returns_dtype=None,
                   remove_infs=None, distinct=False, nulls=False):
        if wrt is None:
            wrt = self._start
        try:
            if predicate:
                n = self._G.add_predicate_aggregation(self._node, wrt._node, string_op)
            else:
                n = self._G.add_aggregation(self._node, wrt._node, string_op, remove_infs, expected_dtype, self.dtype, distinct, nulls)
        except SyntaxError:
            raise DisjointPathError(f"Cannot aggregate {self} into {wrt} since they don't share a parent query")
        from .objects import AttributeQuery
        return AttributeQuery._spawn(self, n, wrt._obj, wrt._node, dtype=returns_dtype, single=True, factor_name=op_name)
